#include <stdio.h>
#include <stdlib.h> /* per utilizzare srand e rand vedi funz. lineacasuale */
#include <time.h>   /* per utilizzare time : vedi funz. lineacasuale */

#define DIMRULE	 8  /* ogni regola corrisponde a un numero di 8 cifre binarie */
#define LDIM	 3  /* le regole hanno effetto su terne di bit */
#define LINEDIM	60  /* dimensione dell'array di celle che contiene l'automa lineare */

#define EMPTY	' '
#define FULL	'*'

/* converte d nella sua rappresentazione binaria nella stringa b, di lunghezza dim */
/* aggiunge tutti gli 0 iniziali necessari a raggiungere la lunghezza dim */
/* cfr. con funz. dec2bin dell'esercizio precedente */
void convert(short d, char *b, short dim)
{
	short i = dim; /* b e' un array di dim + 1 char */

	/* introduco il carattere di terminazione nell'ultima posizione dell'array b */
	b[dim] = 0;
	/* itero trovando una cifra binaria per volta, dalla meno significativa alla */
      	/* piu' significativa; in altri termini, da destra a sinistra */
	do {
		/* sommo '0' per passare dal valore intero 0 o 1 al valore del codice */
                /* ASCII di '0' o '1' */
		b[--i] = (d % 2) + '0'; 
		d /= 2;
	} while(d >= 1);
	/* i adesso e' la posizione iniziale in b della stringa binaria */
	
	/* aggiungo gli zeri iniziali */
	while(--i >= 0)
		b[i] = '0';
}

/* converte una stringa che rappresenta un numero binario, */
/* nel valore short corrispondente */
/* e lo restituisce */
/* cfr. con bin2dec dell'esercizio precedente */
short deconvert(char *b)
{
	short i,c;
	int n = 0;

	/* itero sulle cifre binarie dalla piu' significativa alla meno significativa */
	for(i = 0; b[i]; i++) {
		/* assegno a c il valore corrispondente alla i-esima cifra binaria */
		c = b[i] - '0';
		/* moltiplico *2 il valore corrispondente alle prime i-1 cifre */
		n *= 2;
		/* gli sommo la i-esima cifra */
		n += c;
	}
	return n;
}

/* chiede di immettere il numero di iterazioni */
/* e lo restituisce */
/* non viene effettuato alcun controllo sulla correttezza */
/* esercizio: implementare i controlli di correttezza (cfr. esercizio testinput) */
int legginumeroiterazioni(void)
{
	int i;

	printf("Immetti il numero di iterazioni:");
	scanf("%d",&i);
	return i;
}

/* chiede di immettere il codice della regola */
/* e produce in bcode la codifica binaria della regola immessa */
/* non viene effettuato alcun controllo sulla correttezza */
/* esercizio: implementare i controlli di correttezza (cfr. esercizio testinput) */
void leggicodice(char *bcode)
{
	short code;
	
	printf("Immetti il numero di codice:");
	scanf("%hd",&code);
	convert(code,bcode,DIMRULE);	
}

/* stampa la regola nel formato specificato nel testo dell'esercizio */
void stamparegola(char *bcode)
{
	short i;
	char b[DIMRULE+1];

	for(i = 0;i < DIMRULE;i++) {
		/* converte i numeri da 7 a 0 in binario */
		convert(DIMRULE - i - 1,b,LDIM);
		printf("[%s:%c]",b,bcode[i]);
	}
	putchar('\n');
}

/* genera una linea casuale di bit di dimensioni LINEDIM */
void lineacasuale(char *line)
{
	short i;

	/* setta il valore iniziale del generatore pseudocasuale */
        /* utilizzando il valore dell'orologio interno */
	srand(time(NULL));
	for(i = 0;i < LINEDIM;i++)
		/* rand genera un intero pseudocasualmente */
		/* prendendone il resto della divisione con 2 */
		/* ottengo 0 o 1, pseudocasualmente */
		line[i] = rand() % 2;
}

/* stampa a video lo stato dell'automa line */
void stampalinea(char *line)
{
	short i;

	for(i = 0;i < LINEDIM;i++)
		putchar(line[i] ? FULL : EMPTY);
	putchar('\n');
}

/* calcola la prossima generazione dell'automa line */
/* usando la regola rule */
/* e la memorizza in nextline */
void prossimalinea(char *line, char *nextline, char *rule)
{
	short i,j,pcode;
	char pattern[LDIM+1];

	/* itero su ogni cella dell'automa line */
	for(i = 0;i < LINEDIM;i++) {
		/* itero sui "vicini" della cella line[i] */
		for(j = 0;j < LDIM;j++)
			/* determino la struttura della terna dei vicini di line[i]. */			
			/* Utilizzando il resto della divisione per LINEDIM gestisco anche i casi estremali */
			pattern[j] = line[(i-1+j+LINEDIM)%LINEDIM] + '0';			
		/* inserisco carattere di terminazione per la stringa pattern */
		pattern[LDIM] = 0;
		/* ottengo il valore decimale associato alla terna binaria pattern */
		pcode = deconvert(pattern);
		/* imposto il valore di nextline[i] applicando il valore */
		/* che la regola associa al pattern pcode */
		nextline[i] = rule[DIMRULE - pcode - 1] - '0';	
	}	
}

/* funzione per scambiare i puntatori iniziali */
/* a due array di char (o a due stringhe) */
/* NOTA: devo cambiare il valore di due var di tipo char * */
/* dunque utilizzo puntatori al tipo char *, vale a dire: char ** */
void charpointerswap(char **s, char **t)
{
	char *aux = *s;

	*s = *t;
	*t = aux;	
}

int main(void)
{
	int i;
	char rule[DIMRULE+1];
	/* predispongo due array per contenere la generazione corrente e la successiva */
	char line1[LINEDIM], line2[LINEDIM];
	char *line = line1;  /* line e' la generazione corrente */
	char *nextline = line2; /* nextline e' la generazione successiva */

	i = legginumeroiterazioni();
	leggicodice(rule);
	stamparegola(rule);
	lineacasuale(line);
	stampalinea(line);
	/* itero sul numero delle generazioni impostato dall'utente */
	while(i--) {	
		prossimalinea(line,nextline,rule);
		stampalinea(nextline);
		/* scambio i puntatori, */
		/* in modo che line punti all'array che contiene */
		/* la generazione nuova appena calcolata */
		/* e nextline punti all'altro array */ 
		charpointerswap(&line,&nextline);	
	}
	return 0;
}
