#include <stdio.h>

#define DIM	200
#define FILENAME	"numeri.txt"

/* tipo struttura usato per memorizzare le statistiche relative agli algo di ordinamento */
/* cfrs: numero di confronti fra elementi dell'array */
/* swaps: numero di scambi fra elementi dell'array */
/* assignments: numero di assegnamenti fra elementi dell'array (assegnamenti fra gli indici non sono contati) */
struct counts {
	int cfrs, swaps, assignments;
};

void swap(int *x, int *y)
{
	int aux = *x;

	*x = *y;
	*y = aux;
}

void leggiarray(FILE *fp, int *a, int n)
{
	int i;

	for(i = 0; i < n; i++)
		fscanf(fp,"%d",&a[i]);
}

void stampaarray(int *a, int n)
{
	int i;

	for(i = 0; i < n; i++)
		printf("%d,",a[i]);
	putchar('\n');
}

void copiaarray(int *a, int *b, int n)
{
	int i;

	for(i = 0; i < n; i++)
		b[i] = a[i];
}

/* restituisce un valore di tipo struct counts */
struct counts bubblesort(int *a, int n)
{
	int i, j;
	struct counts c;

	c.cfrs = c.swaps = 0;
	for(i = 0; i < n - 1; i++)
		for(j = n - 1; j > i; j--) {
			c.cfrs++; /* l'istruzione seguente e' un confronto */
			if(a[j-1] > a[j]) {
				c.swaps++; /* l'istruzione seguente e' uno scambio */
				swap(&a[j-1],&a[j]);
			}
		}
	return c;
}

/* restituisce un valore di tipo struct counts */
struct counts selectionsort(int *a, int n)
{
	int i, j;
	struct counts c;

	c.cfrs = c.swaps = 0;
	for(i = 0; i < n - 1; i++)
		for(j = i + 1; j < n; j++) {
			c.cfrs++; /* l'istruzione seguente e' un confronto */
			if(a[j] < a[i]) {
				c.swaps++; /* l'istruzione seguente e' uno scambio */
				swap(&a[j],&a[i]);
			}
		}
	return c;
}

/* restituisce un valore di tipo struct counts */
struct counts selectionsort2(int *a, int n)
{
	int i, j, min;
	struct counts c;

	c.cfrs = c.swaps = 0;
	for(i = 0; i < n - 1; i++) {
		for(min = i, j = i + 1; j < n; j++) {
			c.cfrs++; /* l'istruzione seguente e' un confronto */
			if(a[j] < a[min])
				min = j;
		}
		c.swaps++; /* l'istruzione seguente e' uno scambio */
		swap(&a[i],&a[min]);
	}
	return c;
}

/* restituisce un valore di tipo struct counts */
struct counts insertionsort(int *a, int n)
{
	int i, j;
	struct counts c;

	c.cfrs = c.swaps = 0;
	for(i = 1; i < n; i++) {
		for(j = i; j; j--) {
			c.cfrs++; /* l'istruzione seguente e' un confronto*/
			if(a[j] < a[j-1]) {
				c.swaps++; /* l'istruzione seguente e' uno scambio */
				swap(&a[j-1],&a[j]);
			}
		}
	}
	return c;
}

/* restituisce un valore di tipo struct counts */
struct counts insertionsort2(int *a, int n)
{
	int i, j, k;
	struct counts c;

	c.cfrs = c.swaps = c.assignments = 0;
	for(i = n - 1; i; i--) {
		c.cfrs++; /* l'istruzione seguente e' un confronto */
		if(a[i] < a[i-1]) {
			c.swaps++; /* l'istruzione seguente e' uno scambio */
			swap(&a[i],&a[i-1]);
		}
	}
	for(i = 2; i < n; i++) {
		c.assignments++; /* la prima expr del ciclo for e' costituita da un assegnamento (k = a[i]) */
		c.cfrs++;        /* la seconda expr del ciclo for e' costituita da un confronto (k < a[j-1]) */
		for(j =  i, k = a[i]; k < a[j-1]; j--) {
			c.assignments++; /* per l'assegnamento a[j] = a[j-1] */
			c.cfrs++;        /* per la seconda expr del ciclo for */
			a[j] = a[j-1];
		}
		c.assignments++; /* l'istruzione seguente e' un assegnamento fra elementi dell'array */
		a[j] = k;
	}
	return c;
}

/* restituisce un valore di tipo struct counts */
struct counts insertionsort3(int *a, int n)
{
	int i, j, k, min;
	struct counts c;

	c.cfrs = c.swaps = c.assignments = 0;
 	min = 0;
	for(i = 1; i < n; i++) {
		c.cfrs++;
		if(a[i] < a[min])
			min = i;
 	}
	c.swaps++;
	swap(&a[0],&a[min]);
	for(i = 2; i < n; i++) {
		c.assignments++; /* la prima expr del ciclo for e' costituita da un assegnamento (k = a[i]) */
		c.cfrs++;        /* la seconda expr del ciclo for e' costituita da un confronto (k < a[j-1]) */
		for(j =  i, k = a[i]; k < a[j-1]; j--) {
			c.assignments++; /* per l'assegnamento a[j] = a[j-1] */
			c.cfrs++;        /* per la seconda expr del ciclo for */
			a[j] = a[j-1];
		}
		c.assignments++; /* l'istruzione seguente e' un assegnamento fra elementi dell'array */
		a[j] = k;
	}
	return c;
}


int main(void)
{
	int a[DIM], b[DIM];
	struct counts conti; /* variabile conti usata per memorizzare le statistiche */
	FILE *fp = fopen(FILENAME,"r");

	if(!fp) {
		fprintf(stderr,"Non riesco a leggere il file \"%s\"\n",FILENAME);
		return -1;
	}

	leggiarray(fp,a,DIM);

	copiaarray(a,b,DIM);
	conti = bubblesort(b,DIM);
	stampaarray(b,DIM);
	printf("Bubblesort. Totale confronti:%d. Totale scambi:%d\n",
			conti.cfrs, conti.swaps);

	copiaarray(a,b,DIM);
	conti = selectionsort(b,DIM);
	stampaarray(b,DIM);
	printf("Selectionsort. Totale confronti:%d. Totale scambi:%d\n",
			conti.cfrs, conti.swaps);

	copiaarray(a,b,DIM);
	conti = selectionsort2(b,DIM);
	stampaarray(b,DIM);
	printf("Selectionsort2. Totale confronti:%d. Totale scambi:%d\n",
			conti.cfrs, conti.swaps);

	copiaarray(a,b,DIM);
	conti = insertionsort(b,DIM);
	stampaarray(b,DIM);
	printf("Insertionsort. Totale confronti:%d. Totale scambi:%d\n",
			conti.cfrs, conti.swaps);

	copiaarray(a,b,DIM);
	conti = insertionsort2(b,DIM);
	stampaarray(b,DIM);
	printf("Insertionsort2. Totale confronti:%d. Totale scambi:%d. Totale assegnamenti:%d\n",
			conti.cfrs, conti.swaps, conti.assignments);

	copiaarray(a,b,DIM);
	conti = insertionsort3(b,DIM);
	stampaarray(b,DIM);
	printf("Insertionsort3. Totale confronti:%d. Totale scambi:%d. Totale assegnamenti:%d\n",
			conti.cfrs, conti.swaps, conti.assignments);

	fclose(fp);
	return 0;
}
